(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.FunTask = factory());
}(this, function () { 'use strict';

	function interopDefault(ex) {
		return ex && typeof ex === 'object' && 'default' in ex ? ex['default'] : ex;
	}

	function createCommonjsModule(fn, module) {
		return module = { exports: {} }, fn(module, module.exports), module.exports;
	}

	var index = createCommonjsModule(function (module) {
	module.exports = {
	  equals: 'equals',
	  concat: 'concat',
	  empty: 'empty',
	  map: 'map',
	  ap: 'ap',
	  of: 'of',
	  reduce: 'reduce',
	  sequence: 'sequence',
	  chain: 'chain',
	  extend: 'extend',
	  extract: 'extract',
	  bimap: 'bimap',
	  promap: 'promap'
	}
	});

	interopDefault(index);

	function getImmutableIterator(generator) {
	  var maybeIterator = arguments.length <= 1 || arguments[1] === undefined ? null : arguments[1];
	  var history = arguments.length <= 2 || arguments[2] === undefined ? [] : arguments[2];


	  var iterator = maybeIterator;

	  return function (input) {

	    if (iterator === null) {
	      iterator = generator();
	      history.forEach(function (x) {
	        return iterator.next(x);
	      });
	    }

	    var _iterator$next = iterator.next(input);

	    var value = _iterator$next.value;
	    var done = _iterator$next.done;

	    var next = done ? null : getImmutableIterator(generator, iterator, history.concat([input]));
	    iterator = null;

	    return { value: value, next: next };
	  };
	}

	var runGenerator = function runGenerator(T, generator) {
	  var step = function step(iterator) {
	    return function (prevResult) {
	      var _iterator = iterator(prevResult);

	      var value = _iterator.value;
	      var next = _iterator.next;

	      return next ? T.chain(step(next), value) : value;
	    };
	  };
	  return step(getImmutableIterator(generator))();
	};

	var classCallCheck = function (instance, Constructor) {
	  if (!(instance instanceof Constructor)) {
	    throw new TypeError("Cannot call a class as a function");
	  }
	};

	var createClass = function () {
	  function defineProperties(target, props) {
	    for (var i = 0; i < props.length; i++) {
	      var descriptor = props[i];
	      descriptor.enumerable = descriptor.enumerable || false;
	      descriptor.configurable = true;
	      if ("value" in descriptor) descriptor.writable = true;
	      Object.defineProperty(target, descriptor.key, descriptor);
	    }
	  }

	  return function (Constructor, protoProps, staticProps) {
	    if (protoProps) defineProperties(Constructor.prototype, protoProps);
	    if (staticProps) defineProperties(Constructor, staticProps);
	    return Constructor;
	  };
	}();

	var inherits = function (subClass, superClass) {
	  if (typeof superClass !== "function" && superClass !== null) {
	    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
	  }

	  subClass.prototype = Object.create(superClass && superClass.prototype, {
	    constructor: {
	      value: subClass,
	      enumerable: false,
	      writable: true,
	      configurable: true
	    }
	  });
	  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
	};

	var possibleConstructorReturn = function (self, call) {
	  if (!self) {
	    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
	  }

	  return call && (typeof call === "object" || typeof call === "function") ? call : self;
	};

	var defaultFailureHandler = function defaultFailureHandler(failure) {
	  if (failure instanceof Error) {
	    throw failure;
	  } else {
	    throw new Error('Unhandled task failure: ' + String(failure));
	  }
	};
	var noop = function noop() {};

	var runHelper = function runHelper(body, handlers) {
	  var success = handlers.success;
	  var failure = handlers.failure;
	  var catch_ = handlers.catch;

	  var onCancel = noop;
	  var onClose = noop;
	  var _close = function close() {
	    onClose();
	    // The idea here is to kill links to all stuff that we exposed from runHelper closure.
	    // We expose via the return value (cancelation function) and by passing callbacks to the body.
	    // We reason from an assumption that outer code may keep links to values that we exposed forever.
	    // So we look at all things that referenced in the exposed callbacks and kill them.
	    success = noop;
	    failure = noop;
	    catch_ = noop;
	    onCancel = noop;
	    _close = noop;
	  };
	  var bodyReturn = body(function (x) {
	    var s = success;
	    _close();
	    s(x);
	  }, function (x) {
	    var f = failure;
	    _close();
	    f(x);
	  }, catch_ && function (x) {
	    var c = catch_;
	    _close();
	    c(x);
	  });
	  onCancel = bodyReturn.onCancel || noop;
	  onClose = bodyReturn.onClose || noop;
	  if (_close === noop) {
	    onCancel = noop;
	    onClose();
	  }
	  return function () {
	    onCancel();_close();
	  };
	};

	var Task = function () {
	  function Task() {
	    classCallCheck(this, Task);

	    if (this.constructor === Task) {
	      throw new Error('Don\'t call `new Task()`, call `Task.create()` instead');
	    }
	  }

	  // Creates a task with an arbitrary computation


	  createClass(Task, [{
	    key: 'of',

	    // instance alias for Fantasy Land
	    value: function of(value) {
	      return Task.of(value);
	    }

	    // Creates a task that fails with a given error

	  }, {
	    key: 'empty',

	    // instance alias for Fantasy Land
	    value: function empty() {
	      return Task.empty();
	    }

	    // Given array of tasks creates a task of array

	  }, {
	    key: 'map',
	    value: function map(fn) {
	      return new Map(this, fn);
	    }

	    // Transforms a task by applying `fn` to the failure value

	  }, {
	    key: 'mapRejected',
	    value: function mapRejected(fn) {
	      return new MapRejected(this, fn);
	    }

	    // Transforms a task by applying `sf` to the successful value or `ff` to the failure value

	  }, {
	    key: 'bimap',
	    value: function bimap(ff, fs) {
	      return this.map(fs).mapRejected(ff);
	    }

	    // Transforms a task by applying `fn` to the successful value, where `fn` returns a Task

	  }, {
	    key: 'chain',
	    value: function chain(fn) {
	      return new Chain(this, fn);
	    }

	    // Transforms a task by applying `fn` to the failure value, where `fn` returns a Task

	  }, {
	    key: 'orElse',
	    value: function orElse(fn) {
	      return new OrElse(this, fn);
	    }

	    // Applies the successful value of task `this` to to the successful value of task `otherTask`

	  }, {
	    key: 'ap',
	    value: function ap(otherTask) {
	      return this.chain(function (f) {
	        return otherTask.map(function (x) {
	          return f(x);
	        });
	      });
	    }

	    // Selects the earlier of the two tasks

	  }, {
	    key: 'concat',
	    value: function concat(otherTask) {
	      return Task.race([this, otherTask]);
	    }
	  }, {
	    key: '_run',
	    value: function _run(handlers) {
	      // eslint-disable-line
	      throw new Error('Method run() is not implemented in basic Task class.');
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return '<abstract>';
	    }
	  }, {
	    key: 'toString',
	    value: function toString() {
	      return 'Task.' + this._toString();
	    }
	  }, {
	    key: 'run',
	    value: function run(h) {
	      var handlers = typeof h === 'function' ? { success: h, failure: defaultFailureHandler } : { success: h.success || noop, failure: h.failure || defaultFailureHandler, catch: h.catch };
	      return this._run(handlers);
	    }
	  }, {
	    key: 'runAndLog',
	    value: function runAndLog() {
	      this.run({
	        success: function success(x) {
	          console.log('Success:', x);
	        },
	        // eslint-disable-line
	        failure: function failure(x) {
	          console.log('Failure:', x);
	        }
	      });
	    }
	  }], [{
	    key: 'create',
	    value: function create(computation) {
	      return new FromComputation(computation);
	    }

	    // Creates a task that resolves with a given value

	  }, {
	    key: 'of',
	    value: function of(value) {
	      return new Of(value);
	    }
	  }, {
	    key: 'rejected',
	    value: function rejected(error) {
	      return new Rejected(error);
	    }

	    // Creates a task that never completes

	  }, {
	    key: 'empty',
	    value: function empty() {
	      return new Empty();
	    }
	  }, {
	    key: 'parallel',
	    value: function parallel(tasks) {
	      return new Parallel(tasks);
	    }

	    // Given array of tasks creates a task that completes with the earliest value or error

	  }, {
	    key: 'race',
	    value: function race(task) {
	      return new Race(task);
	    }

	    // Transforms a task by applying `fn` to the successful value

	  }, {
	    key: 'map',
	    value: function map(fn, task) {
	      return new Map(task, fn);
	    }
	  }, {
	    key: 'mapRejected',
	    value: function mapRejected(fn, task) {
	      return new MapRejected(task, fn);
	    }
	  }, {
	    key: 'bimap',
	    value: function bimap(ff, fs, task) {
	      return task.map(fs).mapRejected(ff);
	    }
	  }, {
	    key: 'chain',
	    value: function chain(fn, task) {
	      return new Chain(task, fn);
	    }
	  }, {
	    key: 'orElse',
	    value: function orElse(fn, task) {
	      return new OrElse(task, fn);
	    }
	  }, {
	    key: 'ap',
	    value: function ap(tf, tx) {
	      return tf.chain(function (f) {
	        return tx.map(function (x) {
	          return f(x);
	        });
	      });
	    }
	  }, {
	    key: 'concat',
	    value: function concat(a, b) {
	      return Task.race([a, b]);
	    }
	  }, {
	    key: 'do',
	    value: function _do(generator) {
	      return runGenerator(Task, generator);
	    }
	  }]);
	  return Task;
	}();

	var FromComputation = function (_Task) {
	  inherits(FromComputation, _Task);

	  function FromComputation(computation) {
	    classCallCheck(this, FromComputation);

	    var _this = possibleConstructorReturn(this, Object.getPrototypeOf(FromComputation).call(this));

	    _this._computation = computation;
	    return _this;
	  }

	  createClass(FromComputation, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _computation = this._computation;

	      return runHelper(function (s, f, c) {
	        var cancel = void 0;
	        if (c) {
	          try {
	            cancel = _computation(s, f);
	          } catch (e) {
	            c(e);
	          }
	        } else {
	          cancel = _computation(s, f);
	        }
	        return { onCancel: cancel || noop };
	      }, handlers);
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'create(..)';
	    }
	  }]);
	  return FromComputation;
	}(Task);

	var Of = function (_Task2) {
	  inherits(Of, _Task2);

	  function Of(value) {
	    classCallCheck(this, Of);

	    var _this2 = possibleConstructorReturn(this, Object.getPrototypeOf(Of).call(this));

	    _this2._value = value;
	    return _this2;
	  }

	  createClass(Of, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var success = handlers.success;

	      success(this._value);
	      return noop;
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'of(' + String(this._value) + ')';
	    }
	  }]);
	  return Of;
	}(Task);

	var Rejected = function (_Task3) {
	  inherits(Rejected, _Task3);

	  function Rejected(error) {
	    classCallCheck(this, Rejected);

	    var _this3 = possibleConstructorReturn(this, Object.getPrototypeOf(Rejected).call(this));

	    _this3._error = error;
	    return _this3;
	  }

	  createClass(Rejected, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var failure = handlers.failure;

	      failure(this._error);
	      return noop;
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'rejected(' + String(this._error) + ')';
	    }
	  }]);
	  return Rejected;
	}(Task);

	var Empty = function (_Task4) {
	  inherits(Empty, _Task4);

	  function Empty() {
	    classCallCheck(this, Empty);
	    return possibleConstructorReturn(this, Object.getPrototypeOf(Empty).apply(this, arguments));
	  }

	  createClass(Empty, [{
	    key: 'run',
	    value: function run() {
	      return noop;
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'empty()';
	    }
	  }]);
	  return Empty;
	}(Task);

	var Parallel = function (_Task5) {
	  inherits(Parallel, _Task5);

	  function Parallel(tasks) {
	    classCallCheck(this, Parallel);

	    var _this5 = possibleConstructorReturn(this, Object.getPrototypeOf(Parallel).call(this));

	    _this5._tasks = tasks;
	    return _this5;
	  }

	  createClass(Parallel, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _this6 = this;

	      return runHelper(function (s, f, c) {
	        var length = _this6._tasks.length;
	        var values = Array(length);
	        var completedCount = 0;
	        var runTask = function runTask(task, index) {
	          return task.run({
	            success: function success(x) {
	              values[index] = x;
	              completedCount++;
	              if (completedCount === length) {
	                s(values);
	              }
	            },

	            failure: f,
	            catch: c
	          });
	        };
	        var cancels = _this6._tasks.map(runTask);
	        return {
	          onClose: function onClose() {
	            cancels.forEach(function (cancel) {
	              return cancel();
	            });
	          }
	        };
	      }, handlers);
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'parallel([' + this._tasks.map(function (x) {
	        return x._toString();
	      }).join(', ') + '])';
	    }
	  }]);
	  return Parallel;
	}(Task);

	var Race = function (_Task6) {
	  inherits(Race, _Task6);

	  function Race(tasks) {
	    classCallCheck(this, Race);

	    var _this7 = possibleConstructorReturn(this, Object.getPrototypeOf(Race).call(this));

	    _this7._tasks = tasks;
	    return _this7;
	  }

	  createClass(Race, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _this8 = this;

	      return runHelper(function (success, failure, _catch) {
	        var handlers = { success: success, failure: failure, catch: _catch };
	        var cancels = _this8._tasks.map(function (task) {
	          return task.run(handlers);
	        });
	        return {
	          onClose: function onClose() {
	            cancels.forEach(function (cancel) {
	              return cancel();
	            });
	          }
	        };
	      }, handlers);
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return 'race([' + this._tasks.map(function (x) {
	        return x._toString();
	      }).join(', ') + '])';
	    }
	  }]);
	  return Race;
	}(Task);

	var Map = function (_Task7) {
	  inherits(Map, _Task7);

	  function Map(task, fn) {
	    classCallCheck(this, Map);

	    var _this9 = possibleConstructorReturn(this, Object.getPrototypeOf(Map).call(this));

	    _this9._task = task;
	    _this9._fn = fn;
	    return _this9;
	  }

	  createClass(Map, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _fn = this._fn;
	      var success = handlers.success;
	      var failure = handlers.failure;
	      var catch_ = handlers.catch;

	      return this._task.run({
	        success: function (_success) {
	          function success(_x) {
	            return _success.apply(this, arguments);
	          }

	          success.toString = function () {
	            return _success.toString();
	          };

	          return success;
	        }(function (x) {
	          var value = void 0;
	          if (catch_) {
	            try {
	              value = _fn(x);
	            } catch (e) {
	              catch_(e);
	              return;
	            }
	          } else {
	            value = _fn(x);
	          }
	          success(value);
	        }),

	        failure: failure,
	        catch: catch_
	      });
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return this._task._toString() + '.map(..)';
	    }
	  }]);
	  return Map;
	}(Task);

	var MapRejected = function (_Task8) {
	  inherits(MapRejected, _Task8);

	  function MapRejected(task, fn) {
	    classCallCheck(this, MapRejected);

	    var _this10 = possibleConstructorReturn(this, Object.getPrototypeOf(MapRejected).call(this));

	    _this10._task = task;
	    _this10._fn = fn;
	    return _this10;
	  }

	  createClass(MapRejected, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _fn = this._fn;
	      var success = handlers.success;
	      var failure = handlers.failure;
	      var catch_ = handlers.catch;

	      return this._task.run({
	        success: success,
	        failure: function (_failure) {
	          function failure(_x2) {
	            return _failure.apply(this, arguments);
	          }

	          failure.toString = function () {
	            return _failure.toString();
	          };

	          return failure;
	        }(function (x) {
	          var value = void 0;
	          if (catch_) {
	            try {
	              value = _fn(x);
	            } catch (e) {
	              catch_(e);
	              return;
	            }
	          } else {
	            value = _fn(x);
	          }
	          failure(value);
	        }),

	        catch: catch_
	      });
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return this._task._toString() + '.mapRejected(..)';
	    }
	  }]);
	  return MapRejected;
	}(Task);

	var Chain = function (_Task9) {
	  inherits(Chain, _Task9);

	  function Chain(task, fn) {
	    classCallCheck(this, Chain);

	    var _this11 = possibleConstructorReturn(this, Object.getPrototypeOf(Chain).call(this));

	    _this11._task = task;
	    _this11._fn = fn;
	    return _this11;
	  }

	  createClass(Chain, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _this12 = this;

	      var _fn = this._fn;

	      return runHelper(function (success, failure, catch_) {
	        var cancel2 = noop;
	        var cancel1 = _this12._task.run({
	          success: function (_success2) {
	            function success(_x3) {
	              return _success2.apply(this, arguments);
	            }

	            success.toString = function () {
	              return _success2.toString();
	            };

	            return success;
	          }(function (x) {
	            var spawned = void 0;
	            if (catch_) {
	              try {
	                spawned = _fn(x);
	              } catch (e) {
	                catch_(e);
	              }
	            } else {
	              spawned = _fn(x);
	            }
	            if (spawned) {
	              cancel2 = spawned.run({ success: success, failure: failure, catch: catch_ });
	            }
	          }),

	          failure: failure,
	          catch: catch_
	        });
	        return {
	          onCancel: function onCancel() {
	            cancel1();cancel2();
	          }
	        };
	      }, handlers);
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return this._task._toString() + '.chain(..)';
	    }
	  }]);
	  return Chain;
	}(Task);

	var OrElse = function (_Task10) {
	  inherits(OrElse, _Task10);

	  function OrElse(task, fn) {
	    classCallCheck(this, OrElse);

	    var _this13 = possibleConstructorReturn(this, Object.getPrototypeOf(OrElse).call(this));

	    _this13._task = task;
	    _this13._fn = fn;
	    return _this13;
	  }

	  createClass(OrElse, [{
	    key: '_run',
	    value: function _run(handlers) {
	      var _this14 = this;

	      var _fn = this._fn;

	      return runHelper(function (success, failure, catch_) {
	        var cancel2 = noop;
	        var cancel1 = _this14._task.run({
	          success: success,
	          failure: function (_failure2) {
	            function failure(_x4) {
	              return _failure2.apply(this, arguments);
	            }

	            failure.toString = function () {
	              return _failure2.toString();
	            };

	            return failure;
	          }(function (x) {
	            var spawned = void 0;
	            if (catch_) {
	              try {
	                spawned = _fn(x);
	              } catch (e) {
	                catch_(e);
	              }
	            } else {
	              spawned = _fn(x);
	            }
	            if (spawned) {
	              cancel2 = spawned.run({ success: success, failure: failure, catch: catch_ });
	            }
	          }),

	          catch: catch_
	        });
	        return {
	          onCancel: function onCancel() {
	            cancel1();cancel2();
	          }
	        };
	      }, handlers);
	    }
	  }, {
	    key: '_toString',
	    value: function _toString() {
	      return this._task._toString() + '.orElse(..)';
	    }
	  }]);
	  return OrElse;
	}(Task);

	return Task;

}));